"""
Should be able to use the following import structures (e.g.):
------------------------------------------------------------
import amt_tools
amt_tools.train()
amt_tools.transcribe()
amt_tools.evaluate()
------------------------------------------------------------
from amt_tools import train
train()
------------------------------------------------------------
from amt_tools.train import train
train()
"""

from .evaluate import *
from .train import *
from .transcribe import *

"""
These are necessary in order to be able to access classes
and functions in submodules using only 'import amt_tools'
"""

from . import datasets
from . import features
from . import models
from . import tools



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
import amt_tools.tools as tools

# Regular imports
from abc import abstractmethod
from copy import deepcopy

import numpy as np
import os


def filter_notes_by_duration(pitches, intervals, threshold=0.):
    """
    Remove notes from a collection which have a duration less than a threshold

    Parameters
    ----------
    pitches : ndarray (N)
      Array of pitches corresponding to notes
      N - number of notes
    intervals : ndarray (N x 2)
      Array of onset-offset time pairs corresponding to notes
      N - number of notes
    threshold : float
      Minimum duration (seconds) to keep a note - if set to zero, notes must have non-zero duration

    Returns
    ----------
    pitches : ndarray (N)
      Array of pitches corresponding to notes
      N - number of notes
    intervals : ndarray (N x 2)
      Array of onset-offset time pairs corresponding to notes
      N - number of notes
    """

    # Convert to batched notes for easy indexing
    batched_notes = tools.notes_to_batched_notes(pitches, intervals)
    # Calculate the duration of each note
    durations = batched_notes[:, 1] - batched_notes[:, 0]

    if threshold:
        # Remove notes with duration below the threshold
        batched_notes = batched_notes[durations >= threshold]
    else:
        # Remove zero-duration notes
        batched_notes = batched_notes[durations > threshold]

    # Convert back to loose note groups
    pitches, intervals = tools.batched_notes_to_notes(batched_notes)

    return pitches, intervals


def multi_pitch_to_notes(multi_pitch, times, profile, onsets=None, offsets=None):
    """
    Transcription protocol to convert a multi pitch array into loose MIDI note groups.

    Parameters
    ----------
    multi_pitch : ndarray (F x T)
      Discrete pitch activation map
      F - number of discrete pitches
      T - number of frames
    times : ndarray (N)
      Time in seconds of beginning of each frame
      N - number of time samples (frames)
    profile : InstrumentProfile (instrument.py)
      Instrument profile detailing experimental setup
    onsets : ndarray (F x T) or None (Optional)
      Where to start considering notes "active"
      F - number of discrete pitches
      T - number of frames
    offsets : ndarray (F x T) or None (Optional)
      Where to stop considering notes "active" - currently unused
      F - number of discrete pitches
      T - number of frames

    Returns
    ----------
    pitches : ndarray (N)
      Array of pitches corresponding to notes in MIDI format
      N - number of notes
    intervals : ndarray (N x 2)
      Array of onset-offset time pairs corresponding to notes
      N - number of notes
    """

    if onsets is None:
        # Default the onsets if they were not provided
        onsets = tools.multi_pitch_to_onsets(multi_pitch)

    # Make sure all onsets have corresponding pitch activations
    multi_pitch = np.logical_or(onsets, multi_pitch).astype(tools.FLOAT32)

    # Turn onset activations into impulses at starting frame
    onsets = tools.multi_pitch_to_onsets(onsets)

    # Determine the total number of frames
    num_frames = multi_pitch.shape[-1]

    # Estimate the duration of the track (for bounding note offsets)
    times = np.append(times, times[-1] + tools.estimate_hop_length(times))

    # Create empty lists for note pitches and their time intervals
    pitches, intervals = list(), list()

    # Determine the pitch and frame indices where notes begin
    pitch_idcs, frame_idcs = onsets.nonzero()

    # Loop through note beginnings
    for pitch, frame in zip(pitch_idcs, frame_idcs):
        # Mark onset and start offset counter
        onset, offset = frame, frame + 1

        # Increment the offset counter until one of the following occurs:
        #  1. There are no more frames
        #  2. Pitch is no longer active in the multi pitch array
        #  3. A new onset occurs involving the current pitch
        while True:
            # There are no more frames to count
            maxed_out = offset == num_frames

            if maxed_out:
                # Stop looping
                break

            # There is an activation for the pitch at the next frame
            active_pitch = multi_pitch[pitch, offset]

            if not active_pitch:
                # Stop looping
                break

            # There is an onset for the pitch at the next frame
            new_onset = onsets[pitch, offset]

            if new_onset:
                # Stop looping
                break

            # Include the offset counter
            offset += 1

        # Add the frequency to the list
        pitches.append(pitch + profile.low)

        # Add the interval to the list
        intervals.append([times[onset], times[offset]])

    # Convert the lists to numpy arrays
    pitches, intervals = np.array(pitches), np.array(intervals)

    # Sort notes by onset just for the purpose of being neat
    pitches, intervals = tools.sort_notes(pitches, intervals)

    return pitches, intervals


##################################################
# ESTIMATORS                                     #
##################################################


class ComboEstimator(object):
    """
    A simple wrapper to run multiple estimators in succession.
    Order matters. For instance, a MultiPitchRefiner could be
    chained before a PitchListWrapper to use the refined
    predictions when generating pitch list estimations.
    """

    def __init__(self, estimators):
        """
        Initialize estimators and instantiate.

        Parameters
        ----------
        estimators : list of Estimator
          Estimators to use (in-order) when processing a track
        """

        self.estimators = estimators

    def process_track(self, raw_output, track=None):
        """
        Process the track independently using each estimator.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation
        track : string or None (optional)
          Name of the track to use when writing estimates
        """

        # Copy the raw output dictionary and use it to hold estimates
        output = deepcopy(raw_output)

        # Loop through all of the estimators
        for estimator in self.estimators:
            # Process the track with the estimator and update the estimate dictionary
            output.update(estimator.process_track(output, track))

        return output

    def set_save_dirs(self, save_dir, sub_dirs=None):
        """
        Update the save directories for all of the estimators.

        Parameters
        ----------
        save_dir : string
          Directory under which to write output
        sub_dirs : list of string or None (optional)
          Sub-directories to use underneath 'save_dir' for each estimator
          Specifying None for an individual sub-directory
          will disable saving for the respective estimator
        """

        # Loop through all of the estimators
        for i, estimator in enumerate(self.estimators):
            if sub_dirs is None:
                # Do not add a sub-directory to the path
                new_dir = save_dir
            elif sub_dirs[i] is None:
                # Disable saving for the estimator
                new_dir = None
            else:
                # Append the specified sub-directory if it exists
                new_dir = os.path.join(save_dir, sub_dirs[i])

            # Update the save directory
            estimator.set_save_dir(new_dir)


class Estimator(object):
    """
    Implements a generic music information retrieval estimator.
    """

    def __init__(self, profile, save_dir):
        """
        Initialize parameters common to all estimators and instantiate.

        Parameters
        ----------
        profile : InstrumentProfile (instrument.py)
          Instrument profile detailing experimental setup
        save_dir : string or None (optional)
          Directory where estimates for each track will be written
        """

        self.profile = profile

        self.save_dir = None
        self.set_save_dir(save_dir)

    def set_save_dir(self, save_dir):
        """
        Simple helper function to set and create a new save directory.

        Parameters
        ----------
        save_dir : string or None (optional)
          Directory where estimates for each track will be written
        """

        self.save_dir = save_dir

        if self.save_dir is not None:
            # Create the specified directory if it does not already exist
            os.makedirs(self.save_dir, exist_ok=True)

    @staticmethod
    @abstractmethod
    def get_key():
        """
        Default key describing estimates.
        """

        return NotImplementedError

    @abstractmethod
    def pre_proc(self, raw_output):
        """
        This method can be overridden in order to insert extra steps.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        raw_output : dict
          Copy of parameterized raw output
        """

        # Create a local copy of the output so it is only modified within scope
        raw_output = deepcopy(raw_output)

        return raw_output

    @abstractmethod
    def estimate(self, raw_output):
        """
        Obtain the estimate from the raw output.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation
        """

        return NotImplementedError

    @abstractmethod
    def write(self, estimate, track):
        """
        Specify the protocol for writing the estimates.

        Parameters
        ----------
        estimate : object
          Estimate for a track
        track : string
          Name of the track being processed
        """

        return NotImplementedError

    def process_track(self, raw_output, track=None):
        """
        Combines pre_proc(), estimate(), and write(), and returns output in a dictionary.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation
        track : string or None (optional)
          Name of the track being processed

        Returns
        ----------
        output : dict
          Estimate packed in a dictionary
        """

        # Perform any pre-processing steps
        raw_output = self.pre_proc(raw_output)
        # Obtain estimates for the track
        estimate = self.estimate(raw_output)

        if self.save_dir is not None:
            # Write the results to a text file
            self.write(estimate, track)

        # Return the output in a dictionary
        output = {self.get_key() : estimate}

        return output


class StackedNoteTranscriber(Estimator):
    """
    Estimate stacked notes from stacked multi pitch activation maps.
    """

    def __init__(self, profile, save_dir=None, inhibition_window=None, minimum_duration=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See Estimator class for others...

        inhibition_window : float or None (optional)
          Amount of time after which another note of the same pitch cannot begin
        minimum_duration : float or None (optional)
          Minimum necessary duration to keep a note
        """

        super().__init__(profile, save_dir)

        self.inhibition_window = inhibition_window
        self.minimum_duration = minimum_duration

    @staticmethod
    @abstractmethod
    def get_key():
        """
        Default key for note estimates.
        """

        return tools.KEY_NOTES

    def estimate(self, raw_output):
        """
        Estimate notes for each slice of a stacked multi pitch activation map.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        stacked_notes : dict
          Dictionary containing (slice -> (pitches, intervals)) pairs
        """

        # Obtain the multi pitch activation maps to transcribe
        stacked_multi_pitch = tools.unpack_dict(raw_output, tools.KEY_MULTIPITCH)

        # Determine the number of slices in the stacked multi pitch array
        stack_size = stacked_multi_pitch.shape[-3]

        # Obtain the frame times associated with the activation maps
        times = tools.unpack_dict(raw_output, tools.KEY_TIMES)

        # Obtain the onsets and offsets from the raw output if they exist
        stacked_onsets = tools.unpack_dict(raw_output, tools.KEY_ONSETS)
        stacked_offsets = tools.unpack_dict(raw_output, tools.KEY_OFFSETS)

        # If no onsets were provided, prepare a list of None's
        if stacked_onsets is None:
            stacked_onsets = [None] * stack_size

        # If no offsets were provided, prepare a list of None's
        if stacked_offsets is None:
            stacked_offsets = [None] * stack_size

        # Initialize a dictionary to hold the notes
        stacked_notes = dict()

        # Loop through the slices of the stack
        for slc in range(stack_size):
            # Obtain all of the transcription information for this slice
            multi_pitch, onsets, offsets = stacked_multi_pitch[slc], stacked_onsets[slc], stacked_offsets[slc]

            if self.inhibition_window is not None:
                if onsets is None:
                    # Default the onsets if they were not provided
                    onsets = tools.multi_pitch_to_onsets(multi_pitch)
                    # Remove trailing onsets within inhibition window of a previous onset
                    onsets = tools.inhibit_activations(onsets, times, self.inhibition_window)

            # Transcribe this slice of activations
            pitches, intervals = multi_pitch_to_notes(multi_pitch, times, self.profile, onsets, offsets)

            if self.minimum_duration is not None:
                # Filter the notes by duration
                pitches, intervals = filter_notes_by_duration(pitches, intervals, self.minimum_duration)

            # Add the pitch-interval pairs to the stacked notes dictionary under the slice key
            stacked_notes.update(tools.notes_to_stacked_notes(pitches, intervals, slc))

        return stacked_notes

    def write(self, stacked_notes, track):
        """
        Write slice-wise note estimates to respective files.

        Parameters
        ----------
        stacked_notes : dict
          Dictionary containing (slice -> (pitches, intervals)) pairs
        track : string
          Name of the track being processed
        """

        # Obtain a list of the stacked note keys
        keys = list(stacked_notes.keys())

        # Determine how to name the results
        tag = tools.get_tag(track)

        # Loop through the slices of the stack
        for key in keys:
            # Add another tag for the degree of freedom if more than one
            slice_tag = f'{tag}_{key}' if len(stacked_notes) > 1 else f'{tag}'

            # Construct a path for saving the estimates
            path = os.path.join(self.save_dir, f'{slice_tag}.{tools.TXT_EXT}')

            # Extract the loose note groups from the stack
            pitches, intervals = stacked_notes[key]

            # Write the notes to the path
            tools.write_notes(pitches, intervals, path)


class NoteTranscriber(StackedNoteTranscriber):
    """
    Estimate notes from a multi pitch activation map.
    """

    def __init__(self, profile, save_dir=None, inhibition_window=None, minimum_duration=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See StackedNoteTranscriber class...
        """

        super().__init__(profile, save_dir, inhibition_window, minimum_duration)

    def estimate(self, raw_output):
        """
        Estimate notes from a multi pitch activation map.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        batched_notes : ndarray (N x 3)
          Array of note intervals and pitches by row
          N - number of notes
        """

        # Perform any pre-processing steps
        raw_output = self.pre_proc(raw_output)

        # Obtain the multi pitch activation map to transcribe
        multi_pitch = tools.unpack_dict(raw_output, tools.KEY_MULTIPITCH)

        # Convert the multi pitch array to a stacked multi pitch array
        raw_output[tools.KEY_MULTIPITCH] = tools.multi_pitch_to_stacked_multi_pitch(multi_pitch)

        # Obtain onsets and offsets from output if they exist
        onsets = tools.unpack_dict(raw_output, tools.KEY_ONSETS)
        offsets = tools.unpack_dict(raw_output, tools.KEY_OFFSETS)

        if onsets is not None:
            # Convert onsets to a stacked onset activation map
            raw_output[tools.KEY_ONSETS] = tools.multi_pitch_to_stacked_multi_pitch(onsets)

        if offsets is not None:
            # Convert offsets to a stacked offset activation map
            raw_output[tools.KEY_OFFSETS] = tools.multi_pitch_to_stacked_multi_pitch(offsets)

        # Call the parent class estimate function. Multi pitch is just a special
        # case of stacked multi pitch, where there is only one degree of freedom
        output = super().estimate(raw_output)

        # Add the estimated output to the raw output
        pitches, intervals = tools.stacked_notes_to_notes(output)
        batched_notes = tools.notes_to_batched_notes(pitches, intervals)

        return batched_notes

    def write(self, batched_notes, track):
        """
        Write note estimates to a file.

        Parameters
        ----------
        batched_notes : ndarray (N x 3)
          Array of note intervals and pitches by row
          N - number of notes
        track : string
          Name of the track being processed
        """

        # Convert the batched notes to loose note groups
        pitches, intervals = tools.batched_notes_to_notes(batched_notes)

        # Stack the loose note groups
        stacked_notes = tools.notes_to_stacked_notes(pitches, intervals)

        # Call the parent function
        super().write(stacked_notes, track)


class StackedMultiPitchRefiner(StackedNoteTranscriber):
    """
    Refine stacked multi pitch activation maps, after using them to make note
    predictions, by converting note estimates back into multi pitch activation.
    """

    def __init__(self, profile, save_dir=None, inhibition_window=None, minimum_duration=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See StackedNoteTranscriber class...
        """

        super().__init__(profile, save_dir, inhibition_window, minimum_duration)

    @staticmethod
    @abstractmethod
    def get_key():
        """
        Default key for multi pitch activations.
        """

        return tools.KEY_MULTIPITCH

    def estimate(self, raw_output):
        """
        Refine a stacked multi pitch activation map.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        stacked_multi_pitch : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          S - number of slices in stack
          F - number of discrete pitches
          T - number of frames
        """

        # Attempt to extract pre-existing note estimates
        stacked_notes = tools.unpack_dict(raw_output, tools.KEY_NOTES)

        if stacked_notes is None:
            # Obtain note estimates if they were not provided
            stacked_notes = super().estimate(raw_output)

        # Convert the stacked notes back into stacked multi pitch activation maps
        stacked_multi_pitch = tools.stacked_multi_pitch_to_stacked_onsets(stacked_notes)

        return stacked_multi_pitch

    def write(self, stacked_multi_pitch, track):
        """
        Do nothing. There is no protocol for writing multi pitch activation maps to a file.
        A more appropriate action might be converting them to pitch lists and writing those.

        Parameters
        ----------
        stacked_multi_pitch : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          S - number of slices in stack
          F - number of discrete pitches
          T - number of frames
        track : string
          Name of the track being processed
        """

        pass


class MultiPitchRefiner(NoteTranscriber):
    """
    Refine a multi pitch activation map, after using it to make note
    predictions, by converting note estimates back into multi pitch activation.
    """

    def __init__(self, profile, save_dir=None, inhibition_window=None, minimum_duration=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See StackedNoteTranscriber class...
        """

        super().__init__(profile, save_dir, inhibition_window, minimum_duration)

    @staticmethod
    @abstractmethod
    def get_key():
        """
        Default key for multi pitch activations.
        """

        return tools.KEY_MULTIPITCH

    def estimate(self, raw_output):
        """
        Refine a multi pitch activation map.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        multi_pitch : ndarray (F x T)
          Discrete pitch activation map
          F - number of discrete pitches
          T - number of frames
        """

        # Attempt to extract pre-existing note estimates
        batched_notes = tools.unpack_dict(raw_output, tools.KEY_NOTES)

        if batched_notes is None:
            # Obtain note estimates if they were not provided
            batched_notes = super().estimate(raw_output)

        # Convert the batched notes to loose note groups
        pitches, intervals = tools.batched_notes_to_notes(batched_notes)

        # Obtain the frame times associated with the multi pitch array
        times = tools.unpack_dict(raw_output, tools.KEY_TIMES)

        # Convert the notes back into a multi pitch array
        multi_pitch = tools.notes_to_multi_pitch(pitches, intervals, times, self.profile)

        return multi_pitch

    def write(self, multi_pitch, track):
        """
        Do nothing. There is no protocol for writing multi pitch activation maps to a file.
        A more appropriate action might be converting them to pitch lists and writing those.

        Parameters
        ----------
        multi_pitch : ndarray (F x T)
          Discrete pitch activation map
          F - number of discrete pitches
          T - number of frames
        track : string
          Name of the track being processed
        """

        pass


class StackedPitchListWrapper(Estimator):
    """
    Wrapper for converting stacked multi pitch activations to stacked pitch lists.
    """

    def __init__(self, profile, save_dir=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See Estimator class...
        """

        super().__init__(profile, save_dir)

    @staticmethod
    @abstractmethod
    def get_key():
        """
        Default key for pitch lists.
        """

        return tools.KEY_PITCHLIST

    def estimate(self, raw_output):
        """
        Convert stacked multi pitch activations to stacked pitch lists.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        stacked_pitch_list : dict
          Dictionary containing (slice -> (times, pitch_list)) pairs
        """

        # Obtain the stacked multi pitch activation maps
        stacked_multi_pitch = tools.unpack_dict(raw_output, tools.KEY_MULTIPITCH)

        # Obtain the frame times associated with the stacked activation map
        times = tools.unpack_dict(raw_output, tools.KEY_TIMES)

        # Perform the conversion
        stacked_pitch_list = tools.stacked_multi_pitch_to_stacked_pitch_list(stacked_multi_pitch, times, self.profile)

        return stacked_pitch_list

    def write(self, stacked_pitch_list, track):
        """
        Write slice-wise pitch estimates to respective files.

        Parameters
        ----------
        stacked_pitch_list : dict
          Dictionary containing (slice -> (times, pitch_list)) pairs
        track : string
          Name of the track being processed
        """

        # Obtain a list of the stacked pitch list keys
        keys = list(stacked_pitch_list.keys())

        # Determine how to name the results
        tag = tools.get_tag(track)

        # Loop through the slices of the stack
        for key in keys:
            # Add another tag for the degree of freedom if more than one
            slice_tag = f'{tag}_{key}' if len(stacked_pitch_list) > 1 else f'{tag}'

            # Construct a path for saving the estimates
            path = os.path.join(self.save_dir, f'{slice_tag}.{tools.TXT_EXT}')

            # Extract the pitch list from the stack
            times, pitch_list = stacked_pitch_list[key]

            # Write the notes to the path
            tools.write_pitch_list(times, pitch_list, path)


class PitchListWrapper(StackedPitchListWrapper):
    """
    Wrapper for converting a multi pitch activation map to a pitch lists.
    """

    def __init__(self, profile, save_dir=None):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See Estimator class...
        """

        super().__init__(profile, save_dir)

    def estimate(self, raw_output):
        """
        Convert a multi pitch activation map to a pitch lists.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        times : ndarray (N)
          Time in seconds of beginning of each frame
          N - number of time samples (frames)
        pitch_list : list of ndarray (N x [...])
          Array of pitches corresponding to notes
          N - number of pitch observations (frames)
        """

        # Obtain the multi pitch activation map
        multi_pitch = tools.unpack_dict(raw_output, tools.KEY_MULTIPITCH)

        # Obtain the frame times associated with the activation map
        times = tools.unpack_dict(raw_output, tools.KEY_TIMES)

        # Perform the conversion
        pitch_list = tools.multi_pitch_to_pitch_list(multi_pitch, self.profile)

        return times, pitch_list

    def write(self, pitch_list, track):
        """
        Write pitch estimates to a file.

        Parameters
        ----------
        pitch_list : tuple containing
          times : ndarray (N)
            Time in seconds of beginning of each frame
            N - number of time samples (frames)
          pitch_list : list of ndarray (N x [...])
            Array of pitches corresponding to notes
            N - number of pitch observations (frames)
        track : string
          Name of the track being processed
        """

        # Stack the pitch list
        stacked_pitch_list = tools.pitch_list_to_stacked_pitch_list(*pitch_list)

        # Call the parent function
        super().write(stacked_pitch_list, track)


class TablatureWrapper(Estimator):
    """
    Wrapper for converting tablature to multi pitch.
    """

    def __init__(self, profile, save_dir=None, stacked=False):
        """
        Initialize parameters for the estimator.

        Parameters
        ----------
        See Estimator class...

        stacked : bool
          Whether to collapse into a single representation or leave stacked
        """

        super().__init__(profile, save_dir)

        self.stacked = stacked

    def get_key(self):
        """
        Default key for multi pitch activations.
        """

        return tools.KEY_MULTIPITCH

    def estimate(self, raw_output):
        """
        Convert tablature into a single or stacked multi pitch activation map.

        Parameters
        ----------
        raw_output : dict
          Dictionary containing raw output relevant to estimation

        Returns
        ----------
        multi_pitch : ndarray ((S) x F x T)
          Discrete pitch activation map
          S - number of slices in stack - only if stacked=True
          F - number of discrete pitches
          T - number of frames
        """

        # Obtain the tablature
        tablature = tools.unpack_dict(raw_output, tools.KEY_TABLATURE)

        # Perform the conversion
        multi_pitch = tools.tablature_to_stacked_multi_pitch(tablature, self.profile)

        if not self.stacked:
            multi_pitch = tools.stacked_multi_pitch_to_multi_pitch(multi_pitch)

        return multi_pitch

    def write(self, multi_pitch, track):
        """
        Do nothing. There is no protocol for writing multi pitch activation maps to a file.
        A more appropriate action might be converting them to pitch lists and writing those.

        Parameters
        ----------
        multi_pitch : ndarray ((S) x F x T)
          Discrete pitch activation map
          S - number of slices in stack - only if stacked=True
          F - number of discrete pitches
          T - number of frames
        track : string
          Name of the track being processed
        """

        pass



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
import amt_tools.tools as tools

# Regular imports
from mir_eval.transcription import precision_recall_f1_overlap as evaluate_notes
from mir_eval.multipitch import evaluate as evaluate_frames
from abc import abstractmethod
from scipy.stats import hmean
from copy import deepcopy

import numpy as np
import sys
import os

EPSILON = sys.float_info.epsilon

# TODO - add warning when unpack returns None
# TODO - none of the stacked evaluators have been tested independently
#      - they will likely break during append, average, log, write, etc.

##################################################
# HELPER FUNCTIONS / RESULTS DICTIONARY          #
##################################################


def average_results(results):
    """
    Obtain the average across all tracked results for each metric
    in a results dictionary.

    Parameters
    ----------
    results : dictionary
      Dictionary containing results of tracks arranged by metric

    Returns
    ----------
    average : dictionary
      Dictionary with a single value for each metric
    """

    # Only modify a local copy which will be returned
    average = deepcopy(results)

    # Loop through the keys in the dictionary
    for key in average.keys():
        # Check if the entry is another dictionary
        if isinstance(average[key], dict):
            # Recursively call this function
            average[key] = average_results(average[key])
        else:
            # Check if the entry is a NumPy array or list - leave it alone otherwise
            if isinstance(average[key], np.ndarray) or isinstance(average[key], list):
                # Take the average of all entries and convert to float (necessary for logger)
                average[key] = float(np.mean(average[key]))

    return average


def append_results(tracked_results, new_results):
    """
    Combine two results dictionaries. This function is more general than
    the signature suggests.

    Parameters
    ----------
    tracked_results and new_results : dictionary
      Dictionaries containing results of tracks arranged by metric

    Returns
    ----------
    tracked_results : dictionary
      Dictionary with all results appended along the metric
    """

    # Only modify a local copy which will be returned
    tracked_results = deepcopy(tracked_results)

    # Loop through the keys in the new dictionary
    for key in new_results.keys():
        # Check if the key already exists in the current dictionary
        if key not in tracked_results.keys():
            # Add the untracked entry
            tracked_results[key] = new_results[key]
        # Check if the entry is another dictionary
        elif isinstance(new_results[key], dict):
            # Recursively call this function
            tracked_results[key] = append_results(tracked_results[key], new_results[key])
        else:
            # Append the new entry (or entries) to the current entry
            tracked_results[key] = np.append(tracked_results[key], new_results[key])

    return tracked_results


def log_results(results, writer, step=0, patterns=None, tag=''):
    """
    Log results using TensorBoardX.

    Parameters
    ----------
    results : dictionary
      Dictionary containing results of tracks arranged by metric
    writer : tensorboardX.SummaryWriter
      Writer object being used to log results
    step : int
      Current iteration in whatever process (e.g. training)
    patterns : list of string or None (optional)
      Only write metrics containing these patterns (e.g. ['f1', 'pr']) (None for all metrics)
    tag : string
      Tag for organizing different types of results (e.g. 'validation')
    """

    # Loop through the keys in the dictionary
    for key in results.keys():
        # Extract the next entry
        entry = results[key]

        # Check if the entry is another dictionary
        if isinstance(entry, dict):
            # Add the key to the tag and call this function recursively
            log_results(entry, writer, step, patterns, tag + f'/{key}')
        else:
            # Check if the key matches the specified patterns
            if pattern_match(key, patterns) or patterns is None:
                # Log the entry under the specified key
                writer.add_scalar(f'{tag}/{key}', entry, global_step=step)


def write_results(results, file, patterns=None, verbose=False):
    """
    Write result dictionary to a text file.

    Parameters
    ----------
    results : dictionary
      Dictionary containing results of tracks arranged by metric
    file : TextIOWrapper
      File open in write mode
    patterns : list of string or None (optional)
      Only write metrics containing these patterns (e.g. ['f1', 'pr']) (None for all metrics)
    verbose : bool
      Whether to print to console whatever is written to the file
    """

    # Loop through the keys in the dictionary
    for key in results.keys():
        # Check if the key's entry is another dictionary
        if isinstance(results[key], dict):
            # Write a header to the file
            tools.write_and_print(file, f'-----{key}-----', verbose, '\n')
            # Call this function recursively
            write_results(results[key], file, patterns, verbose)
            # Write an empty line
            tools.write_and_print(file, '', verbose, '\n')
        else:
            # Check if the key matches the specified patterns
            if pattern_match(key, patterns) or patterns is None:
                # Write the metric and corresponding result to the file
                tools.write_and_print(file, f' {key} : {results[key]}', verbose, '\n')

    # Write an empty line
    tools.write_and_print(file, '', verbose, '\n')


def pattern_match(query, patterns=None):
    """
    Simple helper function to see if a query matches a list of strings, even if partially.

    Parameters
    ----------
    query : string
      String to check for matches
    patterns : list of string or None (optional)
      Patterns to reference, return False if unspecified

    Returns
    ----------
    match : bool
      Whether the query matches some pattern, fully or partially
    """

    # Default the returned value
    match = False

    # Check if there are any patterns to analyze
    if patterns is not None:
        # Compare the query to each pattern
        match = any([p in query for p in patterns])

    return match


##################################################
# EVALUATORS                                     #
##################################################


class Evaluator(object):
    """
    Implements a generic music information retrieval evaluator.
    """

    def __init__(self, key, save_dir, patterns, verbose):
        """
        Initialize parameters common to all evaluators and instantiate.

        Parameters
        ----------
        key : string
          Key to use when unpacking data and organizing results
        save_dir : string or None (optional)
          Directory where results for each track will be written
        patterns : list of string or None (optional)
          Only write/log metrics containing these patterns (e.g. ['f1', 'pr']) (None for all metrics)
        verbose : bool
          Whether to print any written text to console as well
        """

        self.key = key

        self.save_dir = None
        self.set_save_dir(save_dir)

        self.patterns = None
        self.set_patterns(patterns)

        self.verbose = None
        self.set_verbose(verbose)

        # Initialize dictionary to track results
        self.results = None
        self.reset_results()

    def set_save_dir(self, save_dir):
        """
        Simple helper function to set and create a new save directory.

        Parameters
        ----------
        save_dir : string or None (optional)
          Directory where estimates for each track will be written
        """

        self.save_dir = save_dir

        if self.save_dir is not None:
            # Create the specified directory if it does not already exist
            os.makedirs(self.save_dir, exist_ok=True)

    def set_patterns(self, patterns):
        """
        Simple helper function to set new patterns.

        Parameters
        ----------
        patterns : list of string or None (optional)
          Only write/log metrics containing these patterns (e.g. ['f1', 'pr']) (None for all metrics)
        """

        self.patterns = patterns

    def set_verbose(self, verbose):
        """
        Simple helper function to set a new verbose flag.

        Parameters
        ----------
        verbose : bool
          Whether to print any written text to console as well
        """

        self.verbose = verbose

    def reset_results(self):
        """
        Reset tracked results to empty dictionary.
        """

        self.results = dict()

    def average_results(self):
        """
        Return the average of the currently tracked results.

        Returns
        ----------
        average : dictionary
          Dictionary with a single value for each metric
        """

        # Average the tracked results
        average = average_results(self.results)

        return average

    def get_key(self):
        """
        Obtain the key being used for the Evaluator.

        Returns
        ----------
        key : string
          Key to use when unpacking data and organizing results
        """

        if self.key is None:
            # Default the key
            key = self.get_default_key()
        else:
            # Use the provided key
            key = self.key

        return key

    @staticmethod
    @abstractmethod
    def get_default_key():
        """
        Provide the default key to use in the event no key was provided.
        """

        return NotImplementedError

    def unpack(self, data):
        """
        Unpack the relevant entry for evaluation if
        a dictionary is provided and the entry exists.

        Parameters
        ----------
        data : object
          Presumably either a dictionary containing ground-truth
          or model output, or the already-unpacked entry

        Returns
        ----------
        data : object
          Unpacked entry or same object provided if no dictionary
        """

        # Determine the relevant key for evaluation
        key = self.get_key()

        # Check if a dictionary was provided and if the key is in the dictionary
        data = tools.try_unpack_dict(data, key)

        return data

    def pre_proc(self, estimated, reference):
        """
        Handle both dictionary input as well as relevant input for
        both estimated and reference data.

        Note: This method can be overridden in order to insert extra steps.

        Parameters
        ----------
        estimated : object
          Dictionary containing ground-truth or the already-unpacked entry
        reference : object
          Dictionary containing model output or the already-unpacked entry

        Returns
        ----------
        estimated : object
          Estimate relevant to the evaluation
        reference : object
          Reference relevant to the evaluation
        """

        # Unpacked estimate and reference if dictionaries were provided
        estimated = self.unpack(estimated)
        reference = self.unpack(reference)

        return estimated, reference

    @abstractmethod
    def evaluate(self, estimated, reference):
        """
        Evaluate an estimate with respect to a reference.

        Parameters
        ----------
        estimated : object
          Estimate relevant to the evaluation or the dictionary containing it
        reference : object
          Reference relevant to the evaluation or the dictionary containing it
        """

        return NotImplementedError

    def write(self, results, track=None):
        """
        Write the results dictionary to a text file if a save directory was specified.

        Parameters
        ----------
        results : dictionary
          Dictionary containing results of tracks arranged by metric
        track : string
          Name of the track being processed
        """

        if self.save_dir is not None:
            # Determine how to name the results
            tag = tools.get_tag(track)

            if self.verbose:
                # Print the track name to console as a header to the results
                print(f'Evaluating track: {tag}')

            # Construct a path for the results
            results_path = os.path.join(self.save_dir, f'{tag}.{tools.TXT_EXT}')

            # Make sure all directories exist (there can be directories in the track name)
            os.makedirs(os.path.dirname(results_path), exist_ok=True)

            # Open a file at the path with writing permissions
            with open(results_path, 'w') as results_file:
                # Write the results to a text file
                write_results(results, results_file, self.patterns, self.verbose)

    def get_track_results(self, estimated, reference, track=None):
        """
        Calculate the results, write them, and track them within the evaluator.

        Parameters
        ----------
        estimated : object
          Estimate relevant to the evaluation or the dictionary containing it
        reference : object
          Reference relevant to the evaluation or the dictionary containing it
        track : string
          Name of the track being processed

        Returns
        ----------
        results : dictionary
          Dictionary containing results of tracks arranged by metric
        """

        # Make sure the estimated and reference data are unpacked
        estimated, reference = self.pre_proc(estimated, reference)

        # Calculate the results
        results = self.evaluate(estimated, reference)

        # Add the results to the tracked dictionary
        self.results = append_results(self.results, results)

        # Write the results
        self.write(results, track)

        return results

    def finalize(self, writer, step=0):
        """
        Log the averaged results using TensorBoardX and reset the results tracking.

        Parameters
        ----------
        writer : tensorboardX.SummaryWriter
          Writer object being used to log results
        step : int
          Current iteration in whatever process (e.g. training)
        """

        # Average the currently tracked results
        average = self.average_results()

        # Log the currently tracked results
        log_results(average, writer, step, patterns=self.patterns, tag=tools.VAL)

        # Reset the tracked results
        self.reset_results()


class ComboEvaluator(Evaluator):
    """
    Packages multiple evaluators into one modules.
    """

    def __init__(self, evaluators, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...

        evaluators : list of Evaluator
          All of the evaluators to run
        """

        self.evaluators = evaluators

        super().__init__(None, save_dir, patterns, verbose)

    def reset_results(self):
        """
        Reset tracked results of each evaluator in the collection.
        """

        # Loop through the evaluators
        for evaluator in self.evaluators:
            # Reset the respective results dictionary so it is empty
            evaluator.reset_results()

    def average_results(self):
        """
        Return the average of the currently tracked results across all evaluators.

        Returns
        ----------
        average : dictionary
          Dictionary with results dictionary entries for each evaluator
        """

        # Initialize an empty dictionary for the average results
        average = dict()

        # Loop through the evaluators
        for evaluator in self.evaluators:
            # Average the tracked results for the evaluator
            # and place in average results under evaluator's key
            results = average_results(evaluator.results)

            # Check if there is already an entry for the evaluator's key
            if tools.query_dict(average, evaluator.get_key()):
                # Add new entries to the results
                average[evaluator.get_key()].update(results)
            else:
                # Create a new entry for the results
                average[evaluator.get_key()] = results

        return average

    @staticmethod
    @abstractmethod
    def get_default_key():
        """
        This should not be called directly on a ComboEvaluator.
        """

        return NotImplementedError

    @abstractmethod
    def evaluate(self, estimated, reference):
        """
        This should not be called directly on a ComboEvaluator.
        """

        return NotImplementedError

    def get_track_results(self, estimated, reference, track=None):
        """
        Very similar to parent method, except file is written after results are
        calculated for each evaluator and packaged into a single dictionary.

        Parameters
        ----------
        estimated : object
          Estimate relevant to the evaluation or the dictionary containing it
        reference : object
          Reference relevant to the evaluation or the dictionary containing it
        track : string
          Name of the track being processed

        Returns
        ----------
        results : dictionary
          Dictionary containing results of tracks arranged by metric
        """

        # Copy the raw output dictionary and use it to hold estimates
        results = {}

        # Loop through the evaluators
        for evaluator in self.evaluators:
            # Make sure the estimated and reference data are unpacked
            estimated_, reference_ = evaluator.pre_proc(estimated, reference)

            # Calculate the results
            new_results = evaluator.evaluate(estimated_, reference_)

            # Check if there is already an entry for the evaluator's key
            if tools.query_dict(results, evaluator.get_key()):
                # Add new entries to the results
                results[evaluator.get_key()].update(new_results)
            else:
                # Create a new entry for the results
                results[evaluator.get_key()] = new_results

            # Add the results to the tracked dictionary
            evaluator.results = append_results(evaluator.results, new_results)

        # Write the results
        self.write(results, track)

        return results


class LossWrapper(Evaluator):
    """
    Simple wrapper for tracking, writing, and logging loss.
    """

    def __init__(self, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...
        """

        super().__init__(key, save_dir, patterns, verbose)

    @staticmethod
    def get_default_key():
        """
        Default key for loss.
        """

        return tools.KEY_LOSS

    def evaluate(self, estimated, reference=None):
        """
        Simply return loss in a new results dictionary.

        Parameters
        ----------
        estimated : ndarray
          Single loss value in a NumPy array
        reference : irrelevant

        Returns
        ----------
        results : dict
          Dictionary containing loss
        """

        # Package the results into a dictionary
        results = estimated

        return results


class StackedMultipitchEvaluator(Evaluator):
    """
    Implements an evaluator for stacked multi pitch activation maps, i.e.
    independent multi pitch estimations across degrees of freedom or instruments.
    """

    def __init__(self, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...
        """

        super().__init__(key, save_dir, patterns, verbose)

    @staticmethod
    def get_default_key():
        """
        Default key for multi pitch activation maps.
        """

        return tools.KEY_MULTIPITCH

    def evaluate(self, estimated, reference):
        """
        Evaluate a stacked multi pitch estimate with respect to a reference.

        Parameters
        ----------
        estimated : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          S - number of slices in stack
          F - number of discrete pitches
          T - number of frames
        reference : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          Dimensions same as estimated

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Determine the shape necessary to flatten the last two dimensions
        flatten_shape = estimated.shape[:-2] + tuple([-1])

        # Flatten the estimated and reference data
        flattened_multi_pitch_est = np.reshape(estimated, flatten_shape)
        flattened_multi_pitch_ref = np.reshape(reference, flatten_shape)

        # Determine the number of correct predictions,
        # where estimated activation lines up with reference
        num_correct = np.sum(flattened_multi_pitch_est * flattened_multi_pitch_ref, axis=-1)

        # Count the number of activations predicted
        num_predicted = np.sum(flattened_multi_pitch_est, axis=-1)
        # Count the number of activations referenced
        num_ground_truth = np.sum(flattened_multi_pitch_ref, axis=-1)

        # Calculate precision and recall
        precision = num_correct / (num_predicted + EPSILON)
        recall = num_correct / (num_ground_truth + EPSILON)

        # Calculate the f1-score using the harmonic mean formula
        f_measure = hmean([precision + EPSILON, recall + EPSILON]) - EPSILON

        # Package the results into a dictionary
        results = {
            tools.KEY_PRECISION : precision,
            tools.KEY_RECALL : recall,
            tools.KEY_F1 : f_measure
        }

        return results


class MultipitchEvaluator(StackedMultipitchEvaluator):
    """
    Implements an evaluator for multi pitch activation maps.
    """

    def __init__(self, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...
        """

        super().__init__(key, save_dir, patterns, verbose)

    def evaluate(self, estimated, reference):
        """
        Evaluate a multi pitch estimate with respect to a reference.

        Parameters
        ----------
        estimated : ndarray (F x T)
          Predicted discrete pitch activation map
          F - number of discrete pitches
          T - number of frames
        reference : ndarray (F x T)
          Ground-truth discrete pitch activation map
          Dimensions same as estimated

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Convert the multi pitch arrays to stacked multi pitch arrays
        stacked_multi_pitch_est = tools.multi_pitch_to_stacked_multi_pitch(estimated)
        stacked_multi_pitch_ref = tools.multi_pitch_to_stacked_multi_pitch(reference)

        # Call the parent class evaluate function. Multi pitch is just a special
        # case of stacked multi pitch, where there is only one degree of freedom
        results = super().evaluate(stacked_multi_pitch_est, stacked_multi_pitch_ref)

        # Average the results across the degree of freedom - i.e. collapse extraneous dimension
        results = average_results(results)

        return results


class StackedNoteEvaluator(Evaluator):
    """
    Implements an evaluator for stacked (independent) note estimations.
    """

    def __init__(self, offset_ratio=None, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...

        offset_ratio : float
          Ratio of the reference note's duration used to define the offset tolerance
        """

        super().__init__(key, save_dir, patterns, verbose)

        self.offset_ratio = offset_ratio

    @staticmethod
    def get_default_key():
        """
        Default key for notes.
        """

        return tools.KEY_NOTES

    def unpack(self, data):
        """
        Unpack notes using the default notes key rather than the specified key.

        Parameters
        ----------
        data : object
          Presumably either a dictionary containing ground-truth
          or model output, or the already-unpacked notes

        Returns
        ----------
        data : object
          Unpacked notes or same object provided if no dictionary
        """

        # Determine the relevant key for evaluation
        key = self.get_default_key()

        # Check if a dictionary was provided and if the key is in the dictionary
        data = tools.try_unpack_dict(data, key)

        return data

    def evaluate(self, estimated, reference):
        """
        Evaluate stacked note estimates with respect to a reference.

        Parameters
        ----------
        estimated : dict
          Dictionary containing (slice -> (pitches, intervals)) pairs
        reference : dict
          Dictionary containing (slice -> (pitches, intervals)) pairs

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Initialize empty arrays to hold results for each degree of freedom
        precision, recall, f_measure = np.empty(0), np.empty(0), np.empty(0)

        # Loop through the stack of notes
        for key in estimated.keys():
            # Extract the loose note groups from the stack
            pitches_ref, intervals_ref = estimated[key]
            pitches_est, intervals_est = reference[key]

            # Convert notes to Hertz
            pitches_ref = tools.notes_to_hz(pitches_ref)
            pitches_est = tools.notes_to_hz(pitches_est)

            # Calculate frame-wise precision, recall, and f1 score with or without offset
            p, r, f, _ = evaluate_notes(ref_intervals=intervals_ref,
                                        ref_pitches=pitches_ref,
                                        est_intervals=intervals_est,
                                        est_pitches=pitches_est,
                                        offset_ratio=self.offset_ratio)

            # Add the results to the respective array
            precision = np.append(precision, p)
            recall = np.append(recall, r)
            f_measure = np.append(f_measure, f)

        # Package the results into a dictionary
        results = {
            tools.KEY_PRECISION : precision,
            tools.KEY_RECALL : recall,
            tools.KEY_F1 : f_measure
        }

        return results


class NoteEvaluator(StackedNoteEvaluator):
    """
    Implements an evaluator for notes.
    """

    def __init__(self, offset_ratio=None, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See StackedNoteEvaluator class...
        """

        super().__init__(offset_ratio, key, save_dir, patterns, verbose)

    def evaluate(self, estimated, reference):
        """
        Evaluate note estimates with respect to a reference.

        Parameters
        ----------
        estimated : ndarray (N x 3)
          Array of estimated note intervals and pitches by row
          N - number of notes
        reference : ndarray (N x 3)
          Array of ground-truth note intervals and pitches by row
          N - number of notes

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Convert the batches notes to notes
        notes_est = tools.batched_notes_to_notes(estimated)
        notes_ref = tools.batched_notes_to_notes(reference)

        # Convert the notes to stacked notes
        stacked_notes_est = tools.notes_to_stacked_notes(*notes_est)
        stacked_notes_ref = tools.notes_to_stacked_notes(*notes_ref)

        # Call the parent class evaluate function
        results = super().evaluate(stacked_notes_est, stacked_notes_ref)

        # Average the results across the degree of freedom - i.e. collapse extraneous dimension
        results = average_results(results)

        return results


class StackedPitchListEvaluator(Evaluator):
    """
    Implements an evaluator for stacked (independent) pitch list estimations.

    This is equivalent to the discrete multi pitch evaluation protocol for
    discrete estimates, but is more general and works for continuous pitch estimations.
    """

    def __init__(self, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...
        """

        super().__init__(key, save_dir, patterns, verbose)

    @staticmethod
    def get_default_key():
        """
        Default key for pitch lists.
        """

        return tools.KEY_PITCHLIST

    def evaluate(self, estimated, reference):
        """
        Evaluate stacked pitch list estimates with respect to a reference.

        Parameters
        ----------
        estimated : dict
          Dictionary containing (slice -> (times, pitch_list)) pairs
        reference : dict
          Dictionary containing (slice -> (times, pitch_list)) pairs

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Initialize empty arrays to hold results for each degree of freedom
        precision, recall, f_measure = np.empty(0), np.empty(0), np.empty(0)

        # Loop through the stack of pitch lists
        for key in estimated.keys():
            # Extract the pitch lists from the stack
            times_ref, pitches_ref = estimated[key]
            times_est, pitches_est = reference[key]

            # Convert pitch lists to Hertz
            pitches_ref = tools.pitch_list_to_hz(pitches_ref)
            pitches_est = tools.pitch_list_to_hz(pitches_est)

            # Calculate frame-wise precision, recall, and f1 score for continuous pitches
            frame_metrics = evaluate_frames(times_ref, pitches_ref, times_est, pitches_est)

            # Extract observation-wise precision and recall
            p, r = frame_metrics['Precision'], frame_metrics['Recall']

            # Calculate the f1-score using the harmonic mean formula
            f = hmean([p + EPSILON, r + EPSILON]) - EPSILON

            # Add the results to the respective array
            precision = np.append(precision, p)
            recall = np.append(recall, r)
            f_measure = np.append(f_measure, f)

        # Package the results into a dictionary
        results = {
            tools.KEY_PRECISION : precision,
            tools.KEY_RECALL : recall,
            tools.KEY_F1 : f_measure
        }

        return results


class PitchListEvaluator(StackedPitchListEvaluator):
    """
    Evaluates pitch list estimates against a reference.

    This is equivalent to the discrete multi pitch evaluation protocol for
    discrete estimates, but is more general and works for continuous pitch estimations.
    """

    def __init__(self, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class...
        """

        super().__init__(key, save_dir, patterns, verbose)

    def evaluate(self, estimated, reference):
        """
        Evaluate pitch list estimates with respect to a reference.

        Parameters
        ----------
        estimated : tuple containing
          times : ndarray (N)
            Time in seconds of beginning of each frame
            N - number of time samples (frames)
          pitch_list : list of ndarray (N x [...])
            Array of pitches corresponding to notes
            N - number of pitch observations (frames)
        reference : tuple containing
          times : ndarray (N)
            Time in seconds of beginning of each frame
            N - number of time samples (frames)
          pitch_list : list of ndarray (N x [...])
            Array of pitches corresponding to notes
            N - number of pitch observations (frames)

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, and f-measure
        """

        # Convert the pitch lists to stacked pitch lists
        stacked_pitch_list_est = tools.pitch_list_to_stacked_pitch_list(*estimated)
        stacked_pitch_list_ref = tools.pitch_list_to_stacked_pitch_list(*reference)

        # Call the parent class evaluate function
        results = super().evaluate(stacked_pitch_list_est, stacked_pitch_list_ref)

        # Average the results across the degree of freedom - i.e. collapse extraneous dimension
        results = average_results(results)

        return results


class TablatureEvaluator(Evaluator):
    """
    Implements an evaluator for tablature.
    """

    def __init__(self, profile, key=None, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class for others...

        profile : InstrumentProfile (instrument.py)
          Instrument profile detailing experimental setup
        """

        super().__init__(key, save_dir, patterns, verbose)

        self.profile = profile

    @staticmethod
    def get_default_key():
        """
        Default key for tablature.
        """

        return tools.KEY_TABLATURE

    def pre_proc(self, estimated, reference):
        """
        By default, we anticipate neither estimate
        or reference to be in stacked multi pitch format.

        TODO - do something similar for pitch list wrapper reference

        Parameters
        ----------
        estimated : object
          Dictionary containing ground-truth or the already-unpacked entry
        reference : object
          Dictionary containing model output or the already-unpacked entry

        Returns
        ----------
        estimated : object
          Estimate relevant to the evaluation
        reference : object
          Reference relevant to the evaluation
        """

        # Unpacked estimate and reference if dictionaries were provided
        tablature_est, tablature_ref = super().pre_proc(estimated, reference)

        # Convert from tablature format to stacked multi pitch format
        tablature_est = tools.tablature_to_stacked_multi_pitch(tablature_est, self.profile)
        tablature_ref = tools.tablature_to_stacked_multi_pitch(tablature_ref, self.profile)

        return tablature_est, tablature_ref

    def evaluate(self, estimated, reference):
        """
        Evaluate a stacked multi pitch tablature estimate with respect to a reference.

        Parameters
        ----------
        estimated : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          S - number of slices in stack
          F - number of discrete pitches
          T - number of frames
        reference : ndarray (S x F x T)
          Array of multiple discrete pitch activation maps
          Dimensions same as estimated

        Returns
        ----------
        results : dict
          Dictionary containing precision, recall, f-measure, and tdr
        """

        # Flatten the estimated and reference data along the pitch and degree-of-freedom axis
        flattened_tablature_est = estimated.flatten()
        flattened_tablature_ref = reference.flatten()

        # Count the number of activations predicted
        num_predicted = np.sum(flattened_tablature_est, axis=-1)
        # Count the number of activations referenced
        num_ground_truth = np.sum(flattened_tablature_ref, axis=-1)

        # Determine the number of correct tablature predictions,
        # where estimated activation lines up with reference
        num_correct_tablature = np.sum(flattened_tablature_est * flattened_tablature_ref, axis=-1)

        # Calculate precision and recall
        precision = num_correct_tablature / (num_predicted + EPSILON)
        recall = num_correct_tablature / (num_ground_truth + EPSILON)

        # Calculate the f1-score using the harmonic mean formula
        f_measure = hmean([precision + EPSILON, recall + EPSILON]) - EPSILON

        # Collapse the stacked multi pitch activations into a single representation
        multi_pitch_est = tools.stacked_multi_pitch_to_multi_pitch(estimated)
        multi_pitch_ref = tools.stacked_multi_pitch_to_multi_pitch(reference)

        # Flatten the estimated and reference multi pitch activations
        flattened_multi_pitch_est = multi_pitch_est.flatten()
        flattened_multi_pitch_ref = multi_pitch_ref.flatten()

        # Determine the number of correct predictions,
        # where estimated activation lines up with reference
        num_correct_multi_pitch = np.sum(flattened_multi_pitch_est * flattened_multi_pitch_ref, axis=-1)

        # Calculate the tablature disambiguation rate
        tdr = num_correct_tablature / (num_correct_multi_pitch + EPSILON)

        # Package the results into a dictionary
        results = {
            tools.KEY_PRECISION : precision,
            tools.KEY_RECALL : recall,
            tools.KEY_F1 : f_measure,
            tools.KEY_TDR : tdr
        }

        return results


class SoftmaxAccuracy(Evaluator):
    """
    Implements an evaluator for calculating accuracy of softmax groups.
    """

    def __init__(self, key, save_dir=None, patterns=None, verbose=False):
        """
        Initialize parameters for the evaluator.

        Parameters
        ----------
        See Evaluator class for others...
        """

        super().__init__(key, save_dir, patterns, verbose)

    @staticmethod
    def get_default_key():
        """
        A key must be provided for softmax groups accuracy.
        """

        return NotImplementedError

    def evaluate(self, estimated, reference):
        """
        Evaluate class membership estimates with respect to a reference.

        Parameters
        ----------
        estimated : ndarray (S x T)
          Array of class membership estimates for multiple degrees of freedom (e.g. strings)
          S - number of degrees of freedom
          T - number of samples or frames
        reference : ndarray (S x F x T)
          Array of class membership ground-truth
          Dimensions same as estimated

        Returns
        ----------
        results : dict
          Dictionary containing accuracy
        """

        # Determine the number of correctly identified classes across all groups
        num_correct = np.sum(estimated == reference)

        # Determine the total number of samples being evaluated
        num_total = reference.size

        # Calculate accuracy averaged over all softmax groups
        accuracy = num_correct / num_total

        # Package the results into a dictionary
        results = {
            tools.KEY_ACCURACY : accuracy
        }

        return results



from setuptools import setup

# TODO - default paths (generated) may be no bueno for usage with pip installation
# TODO - sacred in extras require specifying dependencies for /examples?

setup(
    name='amt-tools',
    url='https://github.com/cwitkowitz/amt-tools',
    author='Frank Cwitkowitz',
    author_email='fcwitkow@ur.rochester.edu',
    packages=['amt_tools'],
    python_requires='>=3.7',
    install_requires=['numpy', 'librosa', 'torch', 'matplotlib',
                      'sacred', 'mir_eval', 'jams', 'mido', 'requests',
                      'tqdm', 'tensorboardX', 'scipy', 'pandas', 'mirdata'],
    #scripts=['examples/of_1.py', 'examples/of_2.py', 'examples/tabcnn.py'],
    version='0.1.1',
    license='MIT',
    description='Machine learning tools and framework for automatic music transcription',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown'
)



# Author: Frank Cwitkowitz <fcwitkow@ur.rochester.edu>

# My imports
from amt_tools import append_results, average_results, log_results

import amt_tools.tools as tools

# Regular imports
from tensorboardX import SummaryWriter
from tqdm import tqdm

import torch.nn as nn
import numpy as np
import torch
import os


def file_sort(file_name):
    """
    Augment file names for sorting within the models directory. Since, e.g.,
    /'500/' will by default be scored as higher than /'1500/'. One way to fix
    this is by adding the length of the file to the beginning of the string.

    Parameters
    ----------
    file_name : str
      Path being sorted

    Returns
    ----------
    sort_name : str
      Character count concatenated with original file name
    """

    # Takes into account the number of digits by adding string length
    sort_name = str(len(file_name)) + file_name

    return sort_name


def validate(model, dataset, evaluator, estimator=None):
    """
    Implements the validation or evaluation loop for a model and dataset partition.
    Optionally save predictions and log results.

    Parameters
    ----------
    model : TranscriptionModel
      Model to validate or evalaute
    dataset : TranscriptionDataset
      Dataset (partition) to use for validation or evaluation
    estimator : Estimator
      Estimation protocol to use
    evaluator : Evaluator
      Evaluation protocol to use

    Returns
    ----------
    average : dict
      Dictionary containing all relevant results averaged across all tracks
    """

    # Make sure the model is in evaluation mode
    model.eval()

    # Turn off gradient computation
    with torch.no_grad():
        # Loop through the validation track ids
        for track_id in dataset.tracks:
            # Obtain the track data
            track = dataset.get_track_data(track_id)

            # Treat the track data as a batch
            batch = tools.track_to_batch(track)

            # Get the model predictions and convert them to NumPy arrays
            predictions = tools.track_to_cpu(model.run_on_batch(batch))

            if estimator is not None:
                # Perform any estimation steps (e.g. note transcription)
                predictions.update(estimator.process_track(predictions, track_id))

            # Evaluate the predictions and track the results
            evaluator.get_track_results(predictions, track, track_id)

    # Obtain the average results from this validation loop
    average = evaluator.average_results()

    return average


def train(model, train_loader, optimizer, iterations, checkpoints=0, log_dir='.', scheduler=None,
          resume=True, single_batch=False, val_set=None, estimator=None, evaluator=None, vis_fnc=None):
    """
    Implements the training loop for an experiment.

    Parameters
    ----------
    model : TranscriptionModel
      Model to train
    train_loader : DataLoader
      PyTorch Dataloader object for retrieving batches of data
    optimizer : Optimizer
      PyTorch Optimizer for updating weights - expected to have only one parameter group
    iterations : int
      Number of loops through the dataset;
      Each loop may be comprised of multiple batches;
      Each loop contains a snippet of each track exactly once
    checkpoints : int
      Number of equally spaced save/validation checkpoints - 0 to disable
    log_dir : str
      Path to directory for saving model, optimizer state, and events
    scheduler : Scheduler or None (optional)
      PyTorch Scheduler used to update learning rate
    resume : bool
      Start from most recently saved model and optimizer state
    single_batch : bool
      Only process the first batch within each validation loop
    val_set : TranscriptionDataset or None (optional)
      Dataset to use for validation loops
    estimator : Estimator
      Estimation protocol to use during validation
    evaluator : Evaluator
      Evaluation protocol to use during validation
    vis_fnc : function(model, i)
      TODO - generalize to any extra validation steps
      Function to perform any visualization steps during validation loop

    Returns
    ----------
    model : TranscriptionModel
      Trained model
    """

    # TODO - multi_gpu - DataParallel removes ability to call .run_on_batch()
    #                  - Can DataParallel be hooked into TranscriptionModel to
    #                    only call the forward() function?

    # Initialize a writer to log any reported results
    writer = SummaryWriter(log_dir)

    # Start at iteration 0 by default
    start_iter = 0

    if resume:
        # Obtain the files that already exist in the log directory
        log_files = os.listdir(log_dir)

        # Extract and sort files pertaining to the model
        model_files = sorted([path for path in log_files if tools.PYT_MODEL in path], key=file_sort)
        # Extract and sort files pertaining to the optimizer state
        optimizer_files = sorted([path for path in log_files if tools.PYT_STATE in path], key=file_sort)

        # Check if any previous checkpoints exist
        if len(model_files) > 0 and len(optimizer_files) > 0:
            # Get the path to the latest model file
            model_path = os.path.join(log_dir, model_files[-1])
            # Get the path to the latest optimizer state file
            optimizer_path = os.path.join(log_dir, optimizer_files[-1])

            # Make the start iteration the iteration when the checkpoint was taken
            start_iter = int(''.join([ch for ch in model_files[-1] if ch.isdigit()]))
            # Get the iteration for the latest optimizer state
            optimizer_iter = int(''.join([ch for ch in optimizer_files[-1] if ch.isdigit()]))
            # Make sure these iterations match
            assert start_iter == optimizer_iter

            # Determine the device to use
            device = model.device
            # Load the latest model and replace the parameterized version
            model = torch.load(model_path, map_location=device)
            model.change_device(device)
            # Replace the randomly initialized parameters with the saved parameters
            # TODO - allow for saving/loading of optimizer with multiple parameter groups
            super(type(optimizer), optimizer).__init__(model.parameters(), optimizer.defaults)
            # Load the latest optimizer state into the parameterized version
            optimizer.load_state_dict(torch.load(optimizer_path, map_location=device))

    # Make sure the model is in training mode
    model.train()

    for global_iter in tqdm(range(start_iter, iterations)):
        # Collection of losses for each batch in the loop
        train_loss = dict()
        # Loop through the dataset
        for batch in train_loader:
            # Zero the accumulated gradients
            optimizer.zero_grad()
            # Get the predictions and loss for the batch
            preds = model.run_on_batch(batch)
            # Extract the loss from the output
            batch_loss = preds[tools.KEY_LOSS]
            # Compute gradients based on total loss
            batch_loss[tools.KEY_LOSS_TOTAL].backward()
            # Add all of the losses to the collection
            train_loss = append_results(train_loss, tools.track_to_cpu(batch_loss))
            # Perform gradient clipping
            # TODO = make optional
            #nn.utils.clip_grad_norm_(model.parameters(), 3)
            # Perform an optimization step
            optimizer.step()

            if single_batch:
                # Move onto the next iteration after the first batch
                break

        if scheduler is not None:
            # Perform a learning rate scheduler step
            scheduler.step()

        # Increase the iteration count by one
        model.iter += 1

        # Average the loss from all of the batches within this loop
        train_loss = average_results(train_loss)
        # Log the training loss(es)
        log_results(train_loss, writer, step=global_iter+1, tag=f'{tools.TRAIN}/{tools.KEY_LOSS}')

        # Local iteration of this training sequence
        local_iter = global_iter - start_iter

        # Set a boolean representing whether current iteration is a checkpoint
        if checkpoints == 0:
            checkpoint = False
        else:
            # TODO - checkpoint at iteration 0?
            checkpoint = (local_iter + 1) % (iterations // checkpoints) == 0

        # Boolean representing whether training has been completed
        done_training = (global_iter + 1) == iterations

        # If we are at a checkpoint, or we have finished training
        if checkpoint or done_training:
            # Save the model
            torch.save(model,
                       os.path.join(log_dir, f'{tools.PYT_MODEL}-{global_iter + 1}.{tools.PYT_EXT}'))
            # Save the optimizer sate
            torch.save(optimizer.state_dict(),
                       os.path.join(log_dir, f'{tools.PYT_STATE}-{global_iter + 1}.{tools.PYT_EXT}'))

            # If visualization protocol was specified, follow it
            if vis_fnc is not None:
                vis_fnc(model, global_iter + 1)

            # If we are at a checkpoint, and a validation set with an estimator is available
            if checkpoint and val_set is not None and evaluator is not None:
                # Validate the current model weights
                validate(model, val_set, evaluator, estimator)
                # Average the results, log them, and reset the tracking
                evaluator.finalize(writer, global_iter + 1)
                # Make sure the model is back in training mode
                model.train()

    return model
